/*
**    FILE    : galois.c   
**    PURPOSE : The functions in this file perform Galois Field arithmetic
**              ( i.e. arithmetic operations in a field with a finite number
**                of elements).
**              These operations are necessary to generate binary and
**              non-binary error-correcting codes.
**    AUTHOR  : Bart De Canne, Belgium (bdc@barco.be)
**    COMPILER: BC++ 3.1 compiler
**    DEPENDENCIES:  Needs "galois.h" as an include file 
**    DATE    : 01-09-1994
**    VERSION : 1.0
*/

#define _GALOIS
#include "galois.h"

static void wr_el_and_optable(FILE *,pGALOIS,pOPTABLE);
static void wr_min_pol(FILE *,pGALOIS);
static void re_el_and_optable(FILE *,pGALOIS,pOPTABLE);
static void re_min_pol(FILE *,pGALOIS);

static int tmp;
static pPOLYNOM pp_tmp;

/* #######################   DYNAMIC ALLOCATION  ######################### */


pPOLYNOM FreePolynom(pPOLYNOM pp) {

   if(pp!=NULL){
      if(pp->c != NULL) {
         free(pp->c);
         pp->c=NULL;
         }
      free(pp);
      pp=NULL;
      }
   return pp;
   }

pPOLYARR FreePolyArr(pPOLYARR pa) {

   int i;

   if(pa!=NULL) {
      for(i=0;i<pa->MaxNo;i++) {
         pa->ppp[i]=FreePolynom(pa->ppp[i]);
         }
      free(pa->ppp);
      free(pa);
      pa=NULL;
      }
   return pa;
   }


pGALOIS FreeGalois(pGALOIS pg) {

   int i;

   if(pg!=NULL) {
      for(i=0;i<pg->MaxNo;i++) {
         pg->pe[i].pp=FreePolynom(pg->pe[i].pp);
         }
      free(pg->pe);
      free(pg);
      pg=NULL;
      }
   return pg;
   }


pOPTABLE FreeOpTable(pOPTABLE po) {

   int i;

   if(po!=NULL) {
      if(po->Add != NULL) {
         for(i=0;i<po->MaxNo;i++) {
            free(po->Add[i]);
            free(po->Mult[i]);
            }
         free(po->Add);
         free(po->Mult);
         free(po->InvAdd);
         free(po->InvMult);
         }
      free(po);
      po=NULL;
      }
   return po;
   }

/* 
** dyn. allocation of a table defining operations in GF(p^q).
** The table is NOT actually constructed.
*/

pOPTABLE AllocOpTable(pOPTABLE po,int p, int q) {

   int i,j;

   BEGIN("AllocOpTable");
   if(po!=NULL)
      po=FreeOpTable(po);

   po=(pOPTABLE) malloc(sizeof(OPTABLE));
   if(po==NULL) ERROR;
   po->Base=p;
   po->Exp=q;
   po->MaxNo=(int)pow((double)p,(double)q);

   po->Add=(byte**)malloc(sizeof(byte*)*po->MaxNo);
   po->Mult=(byte**)malloc(sizeof(byte*)*po->MaxNo);
   po->InvAdd=(byte*)malloc(po->MaxNo);
   po->InvMult=(byte*)malloc(po->MaxNo);

   if(po->Add==NULL || po->Mult==NULL ||
      po->InvAdd==NULL || po->InvMult==NULL)
      ERROR;

   for(i=0;i<po->MaxNo;i++) {
      po->Add[i]=(byte*)malloc(po->MaxNo-i);
      po->Mult[i]=(byte*)malloc(po->MaxNo-i);
      if(po->Add[i]==NULL || po->Mult[i]==NULL) ERROR;
      }
   return po;
   }

/*
** construction of the table <po> given the GF-elements in <pg> and the
** symbol field operations in <op>.
** Note : since operands of + & . operations commute, only half of the
** p^q x p^q table is constructed.
** Hence, when referencing the table of operations,
** the proper entry is found using the I(x,y) and J(x,y) preprocessor
** directives (see galois.h)
** For the GF-indexing convention, see infra.
*/

pOPTABLE ConstructOpTable(pOPTABLE po,pGALOIS pg,pOPTABLE op) {

   int i,j;

   BEGIN("ConstructOpTable");
   if(po==NULL) ERROR;

   START_PACIFIER;
   for(i=0;i<po->MaxNo;i++) {
      PACIFIER(100*i/po->MaxNo);
      po->InvAdd[i]  =  (byte) i_add_inv(i,pg,op);
      po->InvMult[i] =  (byte) i_mult_inv(i,pg,op);
      for(j=i;j<po->MaxNo;j++) {
         po->Add[i][j-i] = (byte) i_add(i,j,pg,op);
         po->Mult[i][j-i]= (byte) i_mult(i,j,pg,op);
         }
      }
   STOP_PACIFIER;
   return po;
   }

/*
**    !!The polynomials in <pg> are NOT allocated
*/

pGALOIS AllocGalois(pGALOIS pg,int p,int q) {

   int i;

   BEGIN("AllocGalois");
   if(pg!=NULL)
      pg=FreeGalois(pg);

   pg=(pGALOIS)malloc(sizeof(GALOIS));
   if(pg==NULL) ERROR;
   pg->Base=p;
   pg->Exp=q;
   pg->MaxNo=(int)pow((double)p,(double)q);
   pg->pe=(pELEMENT)malloc(sizeof(ELEMENT)*pg->MaxNo);
   if(pg->pe==NULL) ERROR;
   for(i=0;i<pg->MaxNo;i++)
      pg->pe[i].pp=NULL;
   return pg;
   }


pPOLYNOM AllocPolynom(pPOLYNOM pp,int p,int max_deg) {

   BEGIN("AllocPolynom");
   if(pp!=NULL)
      pp=FreePolynom(pp);

   pp=(pPOLYNOM) malloc(sizeof(POLYNOM));
   if(pp==NULL) ERROR;
   pp->MaxDeg=max_deg;
   pp->Deg=0;
   pp->Base=p;
   pp->c=(int *)malloc(sizeof(int)*(max_deg+1));
   if(pp->c==NULL) ERROR;
   memset(pp->c,0,sizeof(int)*(max_deg+1));
   return pp;
   }


pPOLYARR AllocPolyArr(pPOLYARR pa,int max_no) {

   BEGIN("AllocPolyArr");
   if(pa!=NULL)
      pa=FreePolyArr(pa);

   pa=(pPOLYARR) malloc(sizeof(POLYARR));
   if(pa==NULL) ERROR;
   pa->MaxNo = max_no;
   pa->FirstFree=0;
   pa->ppp=(ppPOLYNOM) malloc(pa->MaxNo * sizeof(pPOLYNOM));
   if(pa->ppp==NULL) ERROR;
   memset(pa->ppp,0,sizeof(pPOLYNOM)*pa->MaxNo);
   return pa;
   }


/* ######################## CALCULATIONS IN GF ########################### */
/*
**  Note: In the following functions all references to GF-elements denote
**        indexes to the <pg>-structure containing these elements.
**        The convention is:
**           index = 0        ->    GF-element = 0
**                   1        ->                 1
**                   2        ->                 
**                   3        ->                 ^2
**                   ...
**                   p^q - 1  ->                 ^(p^q-2)
*/

/*
** ^k = ^i * ^j
*/

int i_mult(int i,int j,pGALOIS pg,pOPTABLE po) {

   int res;
   int mod=pg->MaxNo;

   if(po==NULL)                  /* symbol field created -> modulo      */
      return( (i*j)%pg->Base);
   if(i==0 || j==0)              /* 0                                   */
      return 0;
   res=i+j-1;
   while(res>=mod)
      res -= (mod-1);
   return res;
   }

/*
** ^k = ^i + ^j
*/

int i_add(int i,int j,pGALOIS pg,pOPTABLE po) {

   byte res[8];
   int k,l;

   if(po==NULL)                  /* symbol field created -> modulo      */
      return( (i+j)%pg->Base);

   for(k=pg->Exp - 1;k>=0;k--) 
      res[k]=po->ADD(pg->pe[i].Alpha[k],pg->pe[j].Alpha[k]);
   for(l=0;;l++) {
      for(k=0;k<pg->Exp;k++) {
         if( res[k] != pg->pe[l].Alpha[k] )
            break;
         }
      if(k==pg->Exp)
         break;
      }
   return l;
   }

/*
** ^k = (^i)^n
*/

int i_pow(int i,int n,pGALOIS pg,pOPTABLE po) {

   int res=1;

   while (n--)
      res=i_mult(res,i,pg,po);
   return res;
   }

/* 
** ^k = - (^i)
*/

int i_add_inv(int i,pGALOIS pg,pOPTABLE po) {

   int j=0;

   while( 0 != i_add(i,j,pg,po) ) j++;
   return j;
   }

/* 
** ^k = (^i)^(-1)
*/

int i_mult_inv(int i,pGALOIS pg,pOPTABLE po) {

   int j=0;

   if(!i) return 0;              /* no multiplicative inverse for 0     */

   while( 1 != i_mult(i,j,pg,po) ) j++;
   return j;
   }


/* ###################### CALCULATIONS USING POLYNOMIALS ################# */

/* 
** Reset polynomial to zero.
*/

void ToZero(pPOLYNOM pp) {

   if(pp==NULL) return;
   memset(pp->c,0,sizeof(int)*(pp->MaxDeg+1));
   pp->Deg=0;
   }

/* 
** Multiplication of polynomials.
** Depending on the value of OP the following function is performed:
** OP==COPY    -> <pp1>*<pp2> is copied into <pp3>
** OP==SUM     -> <pp1>*<pp2> is added to the initial content of <pp3>
** OP==SUBTRACT-> <pp1>*<pp2> is subtracted from the initial content of <pp3>
*/ 

pPOLYNOM MultPoly(pPOLYNOM pp1,pPOLYNOM pp2,pPOLYNOM pp3,pOPTABLE po,
                  byte OP) {

   int i,j;

   BEGIN("MultPoly");
   if(pp1->Base != pp2->Base) ERROR;
   pp3->Deg=pp1->Deg + pp2->Deg;
   if(pp3->MaxDeg < pp3->Deg) ERROR;
                                 /* introduce <pp_tmp> to allow <pp3> to be*/
                                 /* equal to <pp1> or <pp2> in the calling */
                                 /* function                               */
   pp_tmp=AllocPolynom(pp_tmp,pp3->Base,pp3->MaxDeg);
   if(OP!=COPY)                             /* copy it if result has to be */
      pp_tmp=CopyPoly(pp3,pp_tmp);          /* added to or subtracted from */
                                            /* pp3                         */
   for(i=0;i<=pp1->Deg;i++) {
      for(j=0;j<=pp2->Deg;j++) {
         tmp=po->MULT(pp1->c[i],pp2->c[j]);
         if(OP==SUBTRACT)
            tmp=po->InvAdd[tmp];
         pp_tmp->c[i+j]=po->ADD(tmp,pp_tmp->c[i+j]);
         }
      }
   DEGREE(pp_tmp);
   pp3=CopyPoly(pp_tmp,pp3);
   pp_tmp=FreePolynom(pp_tmp);
   return pp3;
   }


/* Division of polynomials : num/denom = quotient + rem/denom
** Returns true if remainder is zero
**              if pp1 is zero
** quot & rem parameters can be specified NULL if only return value of 
** interest
*/

bool DivPoly(pPOLYNOM pp1,pPOLYNOM pp2,pPOLYNOM pp3,pPOLYNOM pp4,
             pOPTABLE po) {

   int *S,L,N,gL,inv_gL,q,i,j,ret_val=1;

   BEGIN("DivPoly");
   L=pp2->Deg;                   /* L=degree of denominator                */
   N=pp1->Deg;                   /* N=degree of numerator                  */
   ToZero(pp3);ToZero(pp4);
   if(pp3) pp3->Deg = N-L;
                                 /* error checking:                        */
   if(N<L) {                     /* 1/ degree of denom. > degree of num.   */
      if(pp3) pp3->Deg=0;
      if(!ISZERO(pp1)) {
         if(pp4)
            pp4=CopyPoly(pp1,pp4);
         return FALSE;
         }
      else
         return TRUE;
      }                          /* 2/ degrees of quot. and remainder      */
   if(pp3)
      if(pp3->MaxDeg < pp3->Deg) ERROR;
   if(pp4)
      if(pp4->MaxDeg < L-1) ERROR;
                                 /* S = shift register of length L         */
   if(NULL==(S=(int*)malloc(L*sizeof(int)))) ERROR;
   memset(S,0,L*sizeof(int));
   gL=pp2->c[L];                 /* highest order coef. of denom. and its  */
   inv_gL=po->InvMult[gL];       /* multiplicative inverse                 */
                                 /* initialization: read L highest order
                                    coeff. of num. into S                  */
   for(i=0;i<L;i++)
      S[i]=pp1->c[N-L+1+i];
                                 /* loop: generate quotient and remainder
                                    by linear feedback shift register, with
                                    feedback coeff. given by denominator   */
   for(i=0;i<=N-L;i++) {
      q=po->MULT(inv_gL,S[L-1]);                /* q = (gL)^(-1)*S[L-1]    */
      if(pp3) pp3->c[N-L-i]=q;
      for(j=L-1;j>=0;j--) {            /* S[j] = S[j-1] - gj*q (j=L-1..1)  */
         tmp=po->MULT(pp2->c[j],q);    /* S[0] = num[N-L-i] - gj*q         */
         tmp=po->InvAdd[tmp];
         S[j]=po->ADD(( (j)?(S[j-1]):(pp1->c[N-L-i]) ),tmp);
         }
      }
   for(i=0;i<L;i++) {
      if(pp4) pp4->c[i]=S[i];
      if(S[i]) ret_val=0;
      }
   if(pp4) DEGREE(pp4);
   if(pp3) DEGREE(pp3);
   free(S);
   return(ret_val);
   }

/* 
** Returns f(i) with f determined by <pp>,
** and <i> the GF-element's index.
*/

int f(pPOLYNOM pp,pGALOIS pg,pOPTABLE po,int i) {

   int k,res=0;

   for(k=pp->Deg;k>=0;k--) {
      if(pp->c[k]==0) continue;
      tmp=i_pow(i,k,pg,po);
      tmp=po->MULT(tmp,pp->c[k]);
      res=po->ADD(tmp,res);
      }
   return res;
   }


/* #########################  FILE I/O  ################################## */

/* 
** Write 'symbol' and 'locator' field to disk.
*/

void WriteGFToDisk(char *file) {

   FILE *fp;

   BEGIN("WriteGFToDisk");
   if((fp = fopen(file,"w+t"))==NULL) ERROR;
   if(po_s->MaxNo != pg_s->MaxNo ||
      po_l->MaxNo != pg_l->MaxNo) ERROR;
   printf("\nWriting...");
   fprintf(fp,"%d %d\n%d %d\n",pg_s->Base,pg_s->Exp,pg_l->Base,pg_l->Exp);
                              /* symbol field: write elements and          */
                              /* table of operations                       */
   wr_el_and_optable(fp,pg_s,po_s);
   fprintf(fp,"$\n");                                             /* token */
                              /* locator field: idem + min.polynomials     */
   wr_el_and_optable(fp,pg_l,po_l);
   wr_min_pol(fp,pg_l);
   printf("OK\n");
   fclose(fp);
   }

void wr_el_and_optable(FILE *fp,pGALOIS pg,pOPTABLE po) {

   int i,j;

   fprintf(fp,"%s\n","FIELD_ELEMENTS:");
   for(i=0;i<pg->MaxNo;i++) {
      for(j=pg->Exp - 1;j>=0;j--)
         fprintf(fp,"%d ",(int)pg->pe[i].Alpha[j]);
      fprintf(fp,"\n");
      }

   fprintf(fp,"%s\n","ADDITION_TABLE:");
   for(i=0;i<po->MaxNo;i++) {
      for(j=i;j<po->MaxNo;j++) {
         fprintf(fp,"%d ",po->Add[i][j-i]);
         }
      fprintf(fp,"\n");
      }

   fprintf(fp,"%s\n","MULTIPLICATION_TABLE:");
   for(i=0;i<po->MaxNo;i++) {
      for(j=i;j<po->MaxNo;j++) {
         fprintf(fp,"%d ",po->Mult[i][j-i]);
         }
      fprintf(fp,"\n");
      }

   fprintf(fp,"%s\n","INVERSES_FOR_ADDITION:");
   for(i=0;i<po->MaxNo;i++) {
      fprintf(fp,"%d ",po->InvAdd[i]);
      }
   fprintf(fp,"\n");

   fprintf(fp,"%s\n","INVERSES_FOR_MULTIPLICATION:");
   for(i=0;i<po->MaxNo;i++) {
      fprintf(fp,"%d ",po->InvMult[i]);
      }
   fprintf(fp,"\n");
   }

void wr_min_pol(FILE *fp,pGALOIS pg) {

   int i,j;

   fprintf(fp,"%s\n","MINIMAL_POLYNOMIALS");
   for(i=2;i<pg->MaxNo;i++) {
      fprintf(fp,"%d: %d -> ",i,pg->pe[i].pp->Deg);
      for(j=pg->pe[i].pp->Deg;j>=0;j--) {
         fprintf(fp,"%d ",pg->pe[i].pp->c[j]);
         }
      fprintf(fp,"\n");
      }
   }

/*
**    Returns <p> & <q> of both 'symbol' and 'locator' field, used by caller
**    to allocate <pg_s>,<po_s> and <pg_l>,<po_l> appropriately.
*/

void ReadGFHeaderFromDisk(char *file,int *p_s,int *q_s,int *p_l,int *q_l) {

   FILE *fp;

   BEGIN("ReadGFHeaderFromDisk");
   if(NULL==(fp=fopen(file,"rt"))) ERROR;
   fscanf(fp,"%d %d\n%d %d\n",p_s,q_s,p_l,q_l);
   fclose(fp);
   }

/* 
** Read 'symbol' and 'locator' field from disk.
** If <pg_s> & <po_s> == NULL    -> only locator field read (el.+op.table)
** If <pg_l> & <po_l> == NULL    -> only symbol field read  (el.+op.table)
** If all <pg_x> & <po_x> != NULL-> read symbol & locator field (incl.pol.)
** 
** !! <pg_x> and <po_x> should be initialized by caller, by calling
**    ReadGFHeaderFromDisk() to get size to allocate.
*/ 

void ReadGFFromDisk( pGALOIS pg_s,pOPTABLE po_s,
                     pGALOIS pg_l,pOPTABLE po_l,
                     char *file) {

   FILE *fp;
   int p_s,q_s,p_l,q_l;
   char c;

   BEGIN("ReadGFFromDisk");
   if(NULL==(fp=fopen(file,"rt"))) ERROR;
   printf("\nReading...");
   fscanf(fp,"%d %d\n%d %d\n",&p_s,&q_s,&p_l,&q_l);

   if(pg_s && po_s) {         /* read symbol field                         */
      if(po_s->MaxNo != pg_s->MaxNo) ERROR;
      if(p_s!=pg_s->Base || q_s!=pg_s->Exp) ERROR;
                              /* symbol field: read elements and           */
                              /* operations table                          */
      re_el_and_optable(fp,pg_s,po_s);
    if(pg_l && po_l) {        /* locator field: idem+min.polynomials       */
         fscanf(fp,"%*s\n");
         re_el_and_optable(fp,pg_l,po_l);
         re_min_pol(fp,pg_l);
         } 
      }
   else {                     /* read only locator field (el.+op.table)    */
      if(po_l->MaxNo != pg_l->MaxNo) ERROR;
      if(p_l!=pg_l->Base || q_l!=pg_l->Exp) ERROR;
                              /* fast forward to locator field             */
      do{
         c=fgetc(fp);
         } while(c!='$');
      c=fgetc(fp);            /* new-line character                        */
      re_el_and_optable(fp,pg_l,po_l);
      }
   printf("OK\n");
   fclose(fp);
   }



void re_el_and_optable(FILE *fp,pGALOIS pg,pOPTABLE po) {

   int i,j;

   fscanf(fp,"%*s\n");
   for(i=0;i<pg->MaxNo;i++) {
      for(j=pg->Exp - 1;j>=0;j--) {
         fscanf(fp,"%d ",&tmp);
         pg->pe[i].Alpha[j] = (byte)tmp;
         }
      fscanf(fp,"\n");
      }
   fscanf(fp,"%*s\n");
   for(i=0;i<po->MaxNo;i++) {
      for(j=i;j<po->MaxNo;j++) {
         fscanf(fp,"%d ",&tmp);
         po->Add[i][j-i] = (byte)tmp;
         }
      fscanf(fp,"\n");
      }
   fscanf(fp,"%*s\n");
   for(i=0;i<po->MaxNo;i++) {
      for(j=i;j<po->MaxNo;j++) {
         fscanf(fp,"%d ",&tmp);
         po->Mult[i][j-i] = (byte)tmp;
         }
      fscanf(fp,"\n");
      }
   fscanf(fp,"%*s\n");
   for(i=0;i<po->MaxNo;i++) {
      fscanf(fp,"%d ",&tmp);
      po->InvAdd[i] = (byte)tmp;
      }
   fscanf(fp,"\n%*s\n");
   for(i=0;i<po->MaxNo;i++) {
      fscanf(fp,"%d ",&tmp);
      po->InvMult[i] = (byte)tmp;
      }
   fscanf(fp,"\n");
   }

void re_min_pol(FILE *fp,pGALOIS pg) {

   int i,j,k,max_deg;

   BEGIN("re_min_pol");
   fscanf(fp,"%*s\n");
   for(i=2;i<pg->MaxNo;i++) {
      fscanf(fp,"%d: %d -> ",&k,&max_deg);
      if(k!=i || max_deg>pg->Exp) ERROR;
      pg->pe[i].pp=AllocPolynom(pg->pe[i].pp,pg->Base,max_deg);
      for(j=max_deg;j>=0;j--) {
         fscanf(fp,"%d ",&tmp);
         pg->pe[i].pp->c[j] = (byte)tmp;
         }
      DEGREE(pg->pe[i].pp);
      fscanf(fp,"\n");
      }
   }



/* ########################### UTILITIES ################################# */

/* 
** Writes <pp> to screen.
*/

void PrintPoly(pPOLYNOM pp) {

   int i;

   for(i=pp->Deg;i>=0;i--) {
      if(pg_l->Exp==1) {
         if(pp->c[i])
            printf("^%d*D^%d ",pp->c[i]-1,i);
         }
      else
         printf("%d*D^%d ",pp->c[i],i);
      }
   printf("\n");
   }

/* 
**  Compares 2 polynomials and returns TRUE if they're equal 
*/

bool ComparePoly(pPOLYNOM pp1,pPOLYNOM pp2) {

   int i;
   bool is_equal=TRUE;

   if(pp1->Deg != pp2->Deg)
      return FALSE;

   for(i=pp1->Deg;i>=0;i--) {
      if(pp1->c[i] != pp2->c[i]) {
         is_equal=FALSE;
         break;
         }
      }
   return is_equal;
   }

/* 
** Copies <pfrom> into <pto> and returns <pto>
*/

pPOLYNOM CopyPoly(pPOLYNOM pfrom,pPOLYNOM pto) {

   int i;

   BEGIN("CopyPoly");
   if(pto->MaxDeg < pfrom->Deg) ERROR;
   ToZero(pto);
   for(i=0;i<=pfrom->Deg;i++) {
      pto->c[i] = pfrom->c[i];
      }
   pto->Deg = pfrom->Deg;
   return pto;
   }

void AskFileName(char *file,char *def) {

   static char tmp[80];

   fflush(stdin);
   printf("\nFilename ? ");
   if(def) 
      printf("<default=%s> : ",def);
   gets(tmp);
   if(def && tmp[0]=='\0')
      strcpy(file,def);
   else
      strcpy(file,tmp);
   }

/* 
** Checks for <n> to be a prime number.
*/ 

bool IsPrime(int n) {

   int i;

   for(i=2;i<n;i++) {
      if(!(n%i))
         return FALSE;
      }
   return TRUE;
   }


/* 
** END OF GALOIS.C
*/ 
